{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE LambdaCase #-}
module Lib
    ( entrypoint,
      -- replLoop
    ) where

import System.IO

import UntypedLambdaExec
import GenericLambda  

import Control.Monad (foldM)
import System.Console.Haskeline
import System.Console.Repline

import Control.Monad.IO.Class (MonadIO(liftIO))
import Control.Monad.State
import Data.List

entrypoint = do
  setupbuffering
  _ <- uLrepl
  return ()


setupbuffering :: IO ()
setupbuffering = do
  hSetBuffering stdin NoBuffering
  hSetBuffering stdout NoBuffering


type Repl a b = HaskelineT (StateT a IO) b

cmd :: (GenericLambda l) => String -> Repl (Env l) () 
cmd input = do
  env <- get
  (toprint, newenv) <- liftIO $ execLine env input
  case toprint of
     Nothing -> pure ()
     Just e -> liftIO $ putStrLn e
  modify (const newenv)

completer :: (GenericLambda l, Monad m, MonadState (Env l) m) => CompletionFunc m
completer = \case
  ("", r) -> do
    matches <- gets (map convertmatch . envStrings)
    return ("", matches)
  (s, r) -> do
    let s' = head $ map reverse $ dumbsep s ""
    matches <- gets (filter (isPrefixOf (reverse s') . fst) . envStrings)
    let completions = map convertmatch matches
    return (s, map (trimComplete s') completions)


  where format s e = s ++ ": " ++ e
        convertmatch (s, e) = Completion s (format s e) True

        dumbsep :: String -> String -> [String]
        dumbsep (' ':rest) curr = curr:dumbsep rest ""
        dumbsep (c:rest) curr = dumbsep rest (c:curr)
        dumbsep "" curr = pure curr

opts = []

ini :: (GenericLambda l) => Repl (Env l) ()  
ini = liftIO $ putStrLn "Type your lambda expression."


final :: (GenericLambda l) => Repl (Env l) ExitDecision
final = do
  liftIO $ putStrLn "Goodbye!"
  return Exit


repl :: (GenericLambda l) => IO ((), Env l)
repl = (runStateT $ evalRepl (const $ pure ">>> ") cmd opts (Just ':') (Just "paste") (Custom completer) ini final) emptyEnv

uLrepl :: IO ((), Env UntypedLambda)
uLrepl = repl


