{-# LANGUAGE GADTs #-}
module TLispAst where

import qualified Data.Map as M

data Pos a = P {filename :: String, row:: Int, col:: Int, obj:: a}


data Literal = LInt Int | Nil | LChar Char | LString String | LBool Bool
  deriving (Eq, Ord)

instance Show Literal where
  show (LInt i) = show i
  show Nil = "nil"
  show (LChar c) = show c
  show (LString s) = s
  show (LBool b) = show b

newtype Id = Id String

instance Show Id where
  show (Id s) = s


-- data Ast where
--   SExp :: Id -> [Ast] -> Ast
--   SLit :: Literal -> Ast
--
newtype Environment = Env (M.Map Id Ast)
data Effect = EPrint String | EError String

data State = S [Effect] Environment

addeff :: State -> Effect -> State
addeff (S es en) e = S (e:es) en

instance Semigroup State where
  (<>) (S es0 env0) (S es1 env1) = S (es0 ++ es1) env1


-- eval :: Environment -> Ast -> ([Effect], Environment, Ast)
-- eval en (SLit a) = ([], en, SLit a)
-- eval en (SExp (Id "print") )


data Ast where
  Car :: Ast -> Ast
  Cdr :: Ast -> Ast
  Cons :: Ast -> Ast -> Ast
  Print :: Ast -> Ast
  Do :: [Ast] -> Ast
  Lit :: Literal -> Ast
  deriving (Eq, Show)

eval :: State -> Ast -> (State, Ast)
eval s (Car a) = case eval s a of
  (s1, Cons b c) -> (s <> s1, b)
  (s1, p) -> (addeff s1 (EError ("Cannot call Car on " ++ show p)), Lit Nil)

eval s (Cdr a) = case eval s a of
  (s1, Cons b c) -> (s <> s1, c)
  (s1, p) -> (addeff s1 (EError ("Cannot call Cdr on " ++ show p)), Lit Nil)

eval s (Print a) = let (s1, k) = eval s a in 
  (addeff s1 (EPrint (show k)), Lit Nil)

eval s (Cons a b) = (s, Cons a b)
eval s x = (s, x)


