{-# LANGUAGE GADTs #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE StandaloneDeriving #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TypeFamilies #-}

module LispyAst where

import qualified Data.Map as M

-- -- we begin representing some primitive types. 


data LType = TFloat
           | TInt
           | TChar
           | TBool
           | TCons LType LType
           | TNil
           | TFun [LType] LType
  deriving (Eq, Ord, Show)

-- -- now, we make some wrappers 
newtype Id = Id String
  deriving (Eq, Ord)

instance Show Id where
  show (Id s) = s



-- instance GEq (TypedId) where
  -- geq (TId i1 t1) (TId i2 t2) = Just ()

-- deriveGEq ''TypedId
-- deriveGCompare ''TypedId
-- deriveArgDict ''TypedId


data Value (a :: LType) where
 LFloat :: Float -> Value TFloat
 -- LInt :: Integer -> Value TInt
 LChar :: Char -> Value TChar
 LBool :: Bool -> Value TBool
 LCons :: Value b -> Value c -> Value (TCons b c)
 Nil :: Value TNil
 LFun :: [Id] -> Value x -> Value (TFun y x)
 -- LVar :: Id -> Value (ToInfer a)
  -- deriving Show
 -- LCar :: (Value (TCons a b)) -> Value a
 -- LCdr :: (Value (TCons a b)) -> Value b
-- there's probably a better way to do this 
-- I mean, there is probably a better way to do this entire mess
-- but it's type safe, ain't that cool

-- instance Eq (Value TFloat) where
--   (==) (LFloat f) (LFloat g) = f == g
--
-- instance Eq (Value TInt) where
--   (==) (LInt a) (LInt b) = a == b
--
instance Num (Value TFloat) where
  (+) (LFloat a) (LFloat b) = LFloat $ a + b
  (*) (LFloat a) (LFloat b) = LFloat $ a * b
  abs (LFloat a) = LFloat $ abs a

instance Show (Value a) where
  show (LFloat f) = show f
  -- show (LInt i) = show i
  show (LChar c) = show c
  show (LCons a b) = "(" ++ show a ++ "," ++ show b ++ ")"
  show (LBool b) = show b
  show Nil = "()"
  show (LFun ids body) = "function: " ++ unwords (map show ids) ++ "."
  -- show (LVar (Id a)) = a 

instance Eq (Value TFloat) where
  (==) (LFloat f) (LFloat g) = f == g
--
instance Ord (Value TFloat) where
  (>=) (LFloat f) (LFloat g) = f >= g
--   (<)
  (<=) (LFloat f) (LFloat g) = f >= g


data Effect = EPrint String | EError String{- |  AddEnv Id [Id] SExp | Lookup Id  -}
  deriving Show



data Ast a where
   IdOp :: Value a -> Ast a
   Exit :: Ast a
   Do :: [WA] -> Ast a -> Ast a

   Print :: (Show (Value v)) => Ast v -> Ast TNil

   Cons :: Ast a -> Ast b -> Ast (TCons a b)
   Car :: Ast (TCons x y) -> Ast x
   Cdr :: Ast (TCons x y) -> Ast y

   -- Lambda :: [Id] -> Value x -> Ast (TFun y x)
   -- Def :: Id -> [Id] -> SExp -> Ast TNil

   DEQ :: Value TFloat -> Value TFloat -> Ast TBool
   DLT :: Value TFloat -> Value TFloat -> Ast TBool
   DGT :: Value TFloat -> Value TFloat -> Ast TBool

   Not :: Ast TBool -> Ast TBool

   Sum :: Ast TFloat -> Ast TFloat -> Ast TFloat
   Sub :: Ast TFloat -> Ast TFloat -> Ast TFloat
   -- Div :: (Fractional (Value b)) => Ast b -> Ast b -> Ast b
   Mul :: Ast TFloat -> Ast TFloat -> Ast TFloat
   Abs :: Ast TFloat -> Ast TFloat
--
deriving instance Show (Value a) => Show (Ast a)
-- deriving instance Num (Value a) => Num (Ast a)
--



interpretdef :: Ast a -> ([Effect], Ast a)
interpretdef (Print a) = ([EPrint (show a)],  IdOp Nil)
--
interpretdef (Car (Cons a b)) = ([], a)
interpretdef (Car a) = let (e1, r1) = interpretdef a
                           (e2, r2) = interpretdef (Car r1)
                        in (e1 ++ e2, r2)
interpretdef (Cdr (Cons a b)) = ([], b)
interpretdef (Cdr a) = let (e1, r1) = interpretdef a
                           (e2, r2) = interpretdef (Cdr r1)
                        in (e1 ++ e2, r2)

interpretdef (Cons a b) = ([],Cons a b)
--
-- interpretdef en (Def id vars body) = ([],en, IdOp Nil)
-- interpretdef en (Lambda vars body) = ([],en, IdOp $ LFun vars body)
--
interpretdef (DEQ x y) = ([], IdOp $ LBool (x == y))
interpretdef (DLT x y) = ([], IdOp $ LBool (x < y))
interpretdef (DGT x y) = ([], IdOp $ LBool (x > y))
--
interpretdef (Not (IdOp (LBool b))) = ([], IdOp $ LBool (not b))
interpretdef (Not a) = let (e1, r1) = interpretdef a
                           (e2, r2) = interpretdef (Not r1)
                        in (e1 ++ e2, r2)
--
interpretdef (Sum (IdOp x) (IdOp y)) = ([], IdOp $ x + y)
interpretdef (Sum a b) = let (e1, r1) = interpretdef a
                             (e2, r2) = interpretdef b
                             (e3, r3) = interpretdef (Sum r1 r2)
                          in (e1 ++ e2, r3)
interpretdef (Sub (IdOp x) (IdOp y)) = ([], IdOp $ x - y)
interpretdef (Sub a b) = let (e1, r1) = interpretdef a
                             (e2, r2) = interpretdef b
                             (e3, r3) = interpretdef (Sub r1 r2)
                          in (e1 ++ e2 ++ e3, r3)
-- interpretdef en (Div x y) = (Nothing, x / y)
interpretdef (Mul (IdOp x) (IdOp y)) = ([], IdOp $ x * y)
interpretdef (Mul a b) = let (e1, r1) = interpretdef a
                             (e2, r2) = interpretdef b
                             (e3, r3) = interpretdef (Mul r1 r2)
                          in (e1 ++ e2 ++ e3, r3)
interpretdef (Abs (IdOp x)) = ([], IdOp $ abs x)

interpretdef (Abs a) = let (e1, r1) = interpretdef a
                           (e2, r2) = interpretdef (Abs r1)
                        in (e1 ++ e2, r2)

interpretdef (IdOp x) = ([], IdOp x)
interpretdef Exit = ([], Exit)
interpretdef (Do [] end) = ([], end)
interpretdef (Do ((WA x):xs) end) = let (e1, _) = interpretdef x
                                        (e2, r2) = interpretdef (Do xs end)
                                in (e1 ++ e2, r2)

--


data SExp where
  -- Builtin :: Ast (Value a) -> SExp
  Literal :: Value a -> SExp
  SExp :: Id -> [SExp] -> SExp
--
instance Show SExp where
  show (Literal v) = show v
  show (SExp id args) = "(" ++ show id ++ " " ++ unwords (map show args) ++ ")"
--
--
-- -- wrapping the value type because existential quantification
-- data WR = forall a. WV (Value a) {- | forall a. WD (Ast (Value a)) -}
data WA = forall a. WA (Ast a)


deriving instance Show WA

-- data WV = forall (a :: LType). WV (Value a)

--
--
--
-- -- convert :: Environment -> SExp -> Maybe WValue
-- -- convert en (Literal v) = Just $ WV v
-- -- convert en  
--
--
data Environment = Env (M.Map Id SExp) Environment | Empty
--
instance Show Environment where
  show _ = "env."
--
-- lookupEnv :: Environment -> TypedId t -> Maybe (Value t)
-- -- lookupEnv = undefined
-- lookupEnv (Env m parent) id = case M.lookup id m of
--   Nothing -> lookupEnv parent id 
--   Just (v) -> Just v
lookupEnv :: Environment -> Id -> Maybe SExp
lookupEnv (Env env parent) id = case M.lookup id env of
  Nothing -> lookupEnv parent id
  Just s -> Just s

lookupEnv Empty _ = Nothing
--
parseSexp :: Environment -> SExp -> Either String (Environment, WA)
parseSexp en (Literal v) = Right (en, WA (IdOp v))
parseSexp en (SExp (Id "car") [a]) = case parseSexp en a of
  Right (en', WA v@(Cons _ _)) -> Right (en', WA (Car v))
  _ -> Left "Car can only be applied to a cons cell."
parseSexp en (SExp (Id "cdr") [a]) = case parseSexp en a of
  Right (en', WA v@(Cons _ _)) -> Right (en', WA (Cdr v))
  _ -> Left "Cdr can only be applied to a cons cell."
parseSexp en (SExp (Id "cons") [a, b]) = case parseSexp en a of
  Right (en', WA v1) -> case parseSexp en' b of
    Right (en'', WA v2) -> Right (en'', WA (Cons v1 v2))
    Left e -> Left e
  Left e -> Left e
-- parseSexp en (Sexp (Id "print") xs) = case 
parseSexp en (SExp (Id "+") [a, b]) = case parseSexp en a of
  Right (en', WA v1@(IdOp (LFloat _))) -> case parseSexp en' b of
    Right (en'', WA v2@(IdOp (LFloat _))) -> Right (en'', WA (Mul v1 v2))
    Right _ -> Left "Argument to + is not a float"
    Left e -> Left e
  Right _ -> Left "Argument to + is not a float"
  Left e -> Left e

-- parseSexp en (SExp (Id "+") (x:xs)) = case parseSexp en x of 
  -- Right (en', WA v1@(IdOp (LInt _))) -> case parseSexp en' (SExp (Id "+") xs) of 
    -- Right (en'', )

  -- Right _ -> Left "Argument to + is not numeric"
  -- Left e -> Left e

parseSexp _ (SExp x _) = Left $ "Wrong number of arguments to " ++ show x
--
--
--
--
-- -- interpret :: WR -> WV  
-- -- interpret (WD a) = let (res, v) = interpretdef en a in interpret (WV v)
-- -- interpret (WV v) = WVF v 
-- -- parseSexp en (SExp id args) = 
--
--
--
--
