{-# LANGUAGE OverloadedStrings #-}
module LispyParser where

import Data.Attoparsec.Text
import LispyAst

import Data.String (fromString)
import Data.Text (unpack)


sexpparser :: Parser SExp
sexpparser = choice [literalparser, parensparser]

identparser :: Parser Id
identparser = Id <$> many1
 (letter <>
  char '+' <>
  char '-' <>
  char '*' <>
  char '_' <>
  char '/')


literalparser = choice [nilparser, falseparser, trueparser, numberparser]

nilparser = do
  string "nil" <> string "()"
  return (Literal Nil)

falseparser = do
  string "false"
  return (Literal $ LBool False)

trueparser = do
  string "true"
  return (Literal $ LBool True)

numberparser = Literal . LFloat . fromIntegral <$> decimal

parensparser :: Parser SExp
parensparser = do
  char '('
  many' space
  id <- identparser
  many' space
  sexps <- sepBy1' sexpparser (many1' space)
  char ')'
  return $ SExp id sexps


runparse = parseOnly sexpparser . fromString

