{-# LANGUAGE TypeFamilies #-}
module UntypedLambdaExec where 

import qualified UntypedLambda as UL
import UL0Parser
import GenericLambda



-- UntypedLambdaCalculus

data UntypedLambda 

instance GenericLambda UntypedLambda where
  data Env UntypedLambda = GEnv UL.Environment
  type Iden UntypedLambda = UL.Identifier
  type Data UntypedLambda = UL.Expression
  data PRes UntypedLambda = ParseResult
  emptyEnv = GEnv UL.emptyEnv
  updateEnv (GEnv e) i d = GEnv $ UL.updateEnv e i d
  envStrings (GEnv e) = UL.envStrings e
  normalform (GEnv e) d = case UL.normalform e d of 
    Nothing -> Left ""
    Just exp -> Right $ show exp
  execLine (GEnv e) s = do 
    (r, e') <- execLine' e s
    return (r, GEnv e')

data ParseResult = Error String | Success | ParsedExp String



evalPrint :: UL.Environment -> UL.Expression -> String
evalPrint en ex = case UL.normalform en ex of 
  Nothing -> "Expression took more than 50 steps before reaching normal form, halting"
  Just res -> "-> " ++ show res


execLine' :: UL.Environment -> String -> IO (Maybe String, UL.Environment)
execLine' en s =
  case runparse s of
    Right (UL.Def id exp) -> return (Nothing, UL.updateEnv en id exp)
    Right (UL.Exp ex)     -> return (Just (evalPrint en ex), en)
    Right (UL.Imp filename) -> do
      env <- parseFile en filename
      return (Nothing, env)
    Left err -> return (Just $ "Parsing failed:" ++ err, en)


-- this function does nothing but calling execLine on a list of statements.
execList :: UL.Environment -> [String] -> IO UL.Environment
execList en [] = return en
execList en (s:ss) = do
  (_, en') <- execLine' en s
  execList en' ss

parseFile :: UL.Environment -> String -> IO UL.Environment
parseFile env filename = do
  content <- readFile filename
  let stmts = filter (/= "") $ lines content
  execList env stmts
